@@ -13,16 +13,15 @@ XPath 2.0 implementation - part 3 (XSD constructors and multi-role tokens)
 import decimal
 import codecs
 from urllib.parse import urlparse
-from urllib.error import URLError
 
 from .exceptions import ElementPathError, xpath_error
-from .xpath_nodes import is_attribute_node
 from .datatypes import DateTime10, Date10, Time, XPathGregorianDay, \
     XPathGregorianMonth, XPathGregorianMonthDay, XPathGregorianYear, \
     XPathGregorianYearMonth, UntypedAtomic, Duration, YearMonthDuration, \
     DayTimeDuration, WHITESPACES_PATTERN, QNAME_PATTERN, NMTOKEN_PATTERN, \
     NAME_PATTERN, NCNAME_PATTERN, HEX_BINARY_PATTERN, NOT_BASE64_BINARY_PATTERN, \
     LANGUAGE_CODE_PATTERN, WRONG_ESCAPE_PATTERN
+from .xpath_context import XPathContext
 from .xpath2_functions import XPath2Parser
 
 
@@ -40,23 +39,16 @@ constructor = XPath2Parser.constructor
 # Constructors for string-based XSD types
 @constructor('normalizedString')
 def cast(value):
-    if isinstance(value, tuple):
-        value = value[-1]
     return str(value).replace('\t', ' ').replace('\n', ' ')
 
 
 @constructor('token')
 def cast(value):
-    if isinstance(value, tuple):
-        value = value[-1]
     return collapse_white_spaces(value)
 
 
 @constructor('language')
 def cast(value):
-    if isinstance(value, tuple):
-        value = value[-1]
-
     match = LANGUAGE_CODE_PATTERN.match(collapse_white_spaces(value))
     if match is None:
         raise xpath_error('FOCA0002', "%r is not a language code" % value)
@@ -65,9 +57,6 @@ def cast(value):
 
 @constructor('NMTOKEN')
 def cast(value):
-    if isinstance(value, tuple):
-        value = value[-1]
-
     match = NMTOKEN_PATTERN.match(collapse_white_spaces(value))
     if match is None:
         raise xpath_error('FOCA0002', "%r is not an xs:NMTOKEN value" % value)
@@ -76,9 +65,6 @@ def cast(value):
 
 @constructor('Name')
 def cast(value):
-    if isinstance(value, tuple):
-        value = value[-1]
-
     match = NAME_PATTERN.match(collapse_white_spaces(value))
     if match is None:
         raise xpath_error('FOCA0002', "%r is not an xs:Name value" % value)
@@ -90,9 +76,6 @@ def cast(value):
 @constructor('IDREF')
 @constructor('ENTITY')
 def cast(value):
-    if isinstance(value, tuple):
-        value = value[-1]
-
     match = NCNAME_PATTERN.match(collapse_white_spaces(value))
     if match is None:
         raise xpath_error('FOCA0002', "invalid value %r for constructor" % value)
@@ -101,14 +84,14 @@ def cast(value):
 
 @constructor('anyURI')
 def cast(value):
-    if isinstance(value, tuple):
-        value = value[-1]
-
     uri = collapse_white_spaces(value)
     try:
-        urlparse(uri)
-    except URLError:
-        raise xpath_error('FOCA0002', "%r is not an xs:anyURI value" % value)
+        url_parts = urlparse(uri)
+        _ = url_parts.port
+    except ValueError as err:
+        msg = "%r is not an xs:anyURI value (%s)"
+        raise xpath_error('FOCA0002', msg % (value, str(err)))
+
     if uri.count('#') > 1:
         msg = "%r is not an xs:anyURI value (too many # characters)"
         raise xpath_error('FOCA0002', msg % value)
@@ -122,8 +105,6 @@ def cast(value):
 # Constructors for numeric XSD types
 @constructor('decimal')
 def cast(value):
-    if isinstance(value, tuple):
-        value = value[-1]
     try:
         return decimal.Decimal(value)
     except (ValueError, decimal.DecimalException) as err:
@@ -133,8 +114,6 @@ def cast(value):
 @constructor('double')
 @constructor('float')
 def cast(value):
-    if isinstance(value, tuple):
-        value = value[-1]
     try:
         return float(value)
     except ValueError as err:
@@ -152,10 +131,7 @@ def cast_to_integer(value, lower_bound=None, higher_bound=None):
     :raise: an `ElementPathValueError` if the value is not decodable to an integer or if \
     the value is out of bounds.
     """
-    if isinstance(value, tuple):
-        value = value[-1]
-
-    if isinstance(value, str):
+    if isinstance(value, (str, bytes)):
         try:
             result = int(float(value))
         except ValueError:
@@ -242,8 +218,6 @@ def cast(value):
 # Constructors for datetime XSD types
 @constructor('date')
 def cast(value, tz=None):
-    if isinstance(value, tuple):
-        value = value[-1]
     if isinstance(value, Date10):
         return value
     return Date10.fromstring(value, tzinfo=tz)
@@ -251,8 +225,6 @@ def cast(value, tz=None):
 
 @constructor('gDay')
 def cast(value, tz=None):
-    if isinstance(value, tuple):
-        value = value[-1]
     if isinstance(value, XPathGregorianDay):
         return value
     return XPathGregorianDay.fromstring(value, tzinfo=tz)
@@ -260,8 +232,6 @@ def cast(value, tz=None):
 
 @constructor('gMonth')
 def cast(value, tz=None):
-    if isinstance(value, tuple):
-        value = value[-1]
     if isinstance(value, XPathGregorianMonth):
         return value
     return XPathGregorianMonth.fromstring(value, tzinfo=tz)
@@ -269,8 +239,6 @@ def cast(value, tz=None):
 
 @constructor('gMonthDay')
 def cast(value, tz=None):
-    if isinstance(value, tuple):
-        value = value[-1]
     if isinstance(value, XPathGregorianMonthDay):
         return value
     return XPathGregorianMonthDay.fromstring(value, tzinfo=tz)
@@ -278,8 +246,6 @@ def cast(value, tz=None):
 
 @constructor('gYear')
 def cast(value, tz=None):
-    if isinstance(value, tuple):
-        value = value[-1]
     if isinstance(value, XPathGregorianYear):
         return value
     return XPathGregorianYear.fromstring(value, tzinfo=tz)
@@ -287,8 +253,6 @@ def cast(value, tz=None):
 
 @constructor('gYearMonth')
 def cast(value, tz=None):
-    if isinstance(value, tuple):
-        value = value[-1]
     if isinstance(value, XPathGregorianYearMonth):
         return value
     return XPathGregorianYearMonth.fromstring(value, tzinfo=tz)
@@ -296,8 +260,6 @@ def cast(value, tz=None):
 
 @constructor('time')
 def cast(value, tz=None):
-    if isinstance(value, tuple):
-        value = value[-1]
     if isinstance(value, Time):
         return value
     return Time.fromstring(value, tzinfo=tz)
@@ -311,11 +273,14 @@ def cast(value, tz=None):
 @method('gYearMonth')
 @method('time')
 def evaluate(self, context=None):
-    item = self.get_argument(context)
-    if item is None:
+    arg = self.data_value(self.get_argument(context))
+    if arg is None:
         return []
+
     try:
-        return self.cast(item, tz=None if context is None else context.timezone)
+        if isinstance(arg, UntypedAtomic):
+            return self.cast(arg.value, tz=None if context is None else context.timezone)
+        return self.cast(arg, tz=None if context is None else context.timezone)
     except ValueError as err:
         raise self.error('FOCA0002', str(err))
     except TypeError as err:
@@ -326,8 +291,6 @@ def evaluate(self, context=None):
 # Constructors for time durations XSD types
 @constructor('duration')
 def cast(value):
-    if isinstance(value, tuple):
-        value = value[-1]
     if isinstance(value, Duration):
         return value
     return Duration.fromstring(value)
@@ -335,8 +298,6 @@ def cast(value):
 
 @constructor('yearMonthDuration')
 def cast(value):
-    if isinstance(value, tuple):
-        value = value[-1]
     if isinstance(value, YearMonthDuration):
         return value
     return YearMonthDuration.fromstring(value)
@@ -344,8 +305,6 @@ def cast(value):
 
 @constructor('dayTimeDuration')
 def cast(value):
-    if isinstance(value, tuple):
-        value = value[-1]
     if isinstance(value, DayTimeDuration):
         return value
     return DayTimeDuration.fromstring(value)
@@ -355,18 +314,14 @@ def cast(value):
 # Constructors for binary XSD types
 @constructor('base64Binary')
 def cast(value, from_literal=False):
-    if isinstance(value, tuple):
-        value = value[-1]
-    if isinstance(value, UntypedAtomic):
-        return codecs.encode(str(value), 'base64')
-    elif not isinstance(value, (bytes, str)):
+    if not isinstance(value, (bytes, str)):
         raise xpath_error('FORG0006', 'the argument has an invalid type %r' % type(value))
     elif not isinstance(value, bytes) or from_literal:
         return codecs.encode(value.encode('ascii'), 'base64')
     elif HEX_BINARY_PATTERN.search(value.decode('utf-8')):
         value = codecs.decode(value, 'hex') if str is not bytes else value
         return codecs.encode(value, 'base64')
-    elif NOT_BASE64_BINARY_PATTERN.search(value.decode('utf-8')):
+    elif NOT_BASE64_BINARY_PATTERN.search(value.decode('utf-8')) or len(value.strip()) % 4:
         return codecs.encode(value, 'base64')
     else:
         return value
@@ -374,19 +329,12 @@ def cast(value, from_literal=False):
 
 @constructor('hexBinary')
 def cast(value, from_literal=False):
-    if isinstance(value, tuple):
-        value = value[-1]
-    if isinstance(value, UntypedAtomic):
-        return codecs.encode(str(value), 'hex')
-    elif not isinstance(value, (bytes, str)):
+    if not isinstance(value, (bytes, str)):
         raise xpath_error('FORG0006', 'the argument has an invalid type %r' % type(value))
     elif not isinstance(value, bytes) or from_literal:
         return codecs.encode(value.encode('ascii'), 'hex')
     elif HEX_BINARY_PATTERN.search(value.decode('utf-8')):
-        if isinstance(value, bytes) or str is bytes:
-            return value
-        else:
-            return codecs.encode(value.encode('ascii'), 'hex')
+        return value
     else:
         try:
             value = codecs.decode(value, 'base64')
@@ -399,89 +347,31 @@ def cast(value, from_literal=False):
 @method('base64Binary')
 @method('hexBinary')
 def evaluate(self, context=None):
-    item = self.get_argument(context)
-    if item is None:
+    arg = self.data_value(self.get_argument(context))
+    if arg is None:
         return []
+
     try:
-        return self.cast(item, self[0].label == 'literal')
+        if isinstance(arg, UntypedAtomic):
+            return self.cast(arg.value, self[0].label == 'literal')
+        return self.cast(arg, self[0].label == 'literal')
     except ElementPathError as err:
-        if err.token is None:
-            err.token = self
+        err.token = self
         raise
-    except ValueError as err:
-        raise self.error('FOCA0002', str(err))
-    except TypeError as err:
-        raise self.error('FORG0006', str(err))
 
 
 ###
-# Multi role-tokens cases
+# Multi role-tokens constructors
 #
 
-# Case 1: In XPath 2.0 the 'attribute' keyword is used both for attribute:: axis and
-# attribute() node type function.
-#
-# First the XPath1 token class has to be removed from the XPath2 symbol table. Then the
-# symbol has to be registered usually with the same binding power (bp --> lbp, rbp), a
-# multi-value label (using a tuple of values) and a custom pattern. Finally a custom nud
-# or led method is required.
-unregister('attribute')
-register('attribute', lbp=90, rbp=90, label=('kind test', 'axis'),
-         pattern=r'\battribute(?=\s*\:\:|\s*\(\:.*\:\)\s*\:\:|\s*\(|\s*\(\:.*\:\)\()')
-
-
-@method('attribute')
-def nud(self):
-    if self.parser.next_token.symbol == '::':
-        self.parser.advance('::')
-        self.parser.next_token.expected(
-            '(name)', '*', 'text', 'node', 'document-node', 'comment', 'processing-instruction',
-            'attribute', 'schema-attribute', 'element', 'schema-element'
-        )
-        self[:] = self.parser.expression(rbp=90),
-        self.label = 'axis'
-    else:
-        self.parser.advance('(')
-        if self.parser.next_token.symbol != ')':
-            self[:] = self.parser.expression(5),
-            if self.parser.next_token.symbol == ',':
-                self.parser.advance(',')
-                self[1:] = self.parser.expression(5),
-        self.parser.advance(')')
-        self.label = 'kind test'
-    return self
-
-
-@method('attribute')
-def select(self, context=None):
-    if context is None:
-        return
-    elif self.label == 'axis':
-        for _ in context.iter_attributes():
-            yield from self[0].select(context)
-    else:
-        name = self[0].evaluate(context) if self else None
-        for item in context.iter_attributes():
-            if is_attribute_node(item, name):
-                yield context.item[1]
-
-
-@method('attribute')
-def evaluate(self, context=None):
-    if context is not None:
-        if is_attribute_node(context.item, self[0].evaluate(context) if self else None):
-            return context.item[1]
-
-
-# Case 2: In XPath 2.0 the 'boolean' keyword is used both for boolean() function and
+# Case 1: In XPath 2.0 the 'boolean' keyword is used both for boolean() function and
 # for boolean() constructor.
 def cast_to_boolean(value, context=None):
+    assert context is None or isinstance(context, XPathContext)
     if isinstance(value, bool):
         return value
     elif isinstance(value, (int, float, decimal.Decimal)):
         return bool(value)
-    elif isinstance(value, UntypedAtomic):
-        value = str(value)
     elif not isinstance(value, str):
         raise xpath_error('FORG0006', 'the argument has an invalid type %r' % type(value))
 
@@ -514,23 +404,27 @@ def nud(self):
 @method('boolean')
 def evaluate(self, context=None):
     if self.label == 'function':
-        return self.boolean_value(self[0].get_results(context))
+        return self.boolean_value([x for x in self[0].select(context)])
 
     # xs:boolean constructor
-    item = self.get_argument(context)
-    if item is None:
+    arg = self.data_value(self.get_argument(context))
+    if arg is None:
         return []
+
     try:
-        return self.cast(item, context)
+        if isinstance(arg, UntypedAtomic):
+            return self.cast(arg.value, context)
+        return self.cast(arg, context)
     except ElementPathError as err:
         err.token = self
         raise
 
 
-# Case 3: In XPath 2.0 the 'string' keyword is used both for fn:string() and xs:string().
+# Case 2: In XPath 2.0 the 'string' keyword is used both for fn:string() and xs:string().
 unregister('string')
-register('string', lbp=90, rbp=90, label=('function', 'constructor'),
-         pattern=r'\bstring(?=\s*\(|\s*\(\:.*\:\)\()', cast=staticmethod(lambda v, c=None: str(v)))
+register('string', lbp=90, rbp=90, label=('function', 'constructor'),  # pragma: no cover
+         pattern=r'\bstring(?=\s*\(|\s*\(\:.*\:\)\()',
+         cast=staticmethod(lambda v, c=None: str(v)))
 
 
 @method('string')
@@ -548,10 +442,10 @@ def evaluate(self, context=None):
         return self.string_value(self.get_argument(context))
     else:
         item = self.get_argument(context)
-        return [] if item is None else str(item)
+        return [] if item is None else self.string_value(item)
 
 
-# Case 4 and 5: In XPath 2.0 the XSD 'QName' and 'dateTime' types have special
+# Case 3 and 4: In XPath 2.0 the XSD 'QName' and 'dateTime' types have special
 # constructor functions so the 'QName' keyword is used both for fn:QName() and
 # xs:QName(), the same for 'dateTime' keyword.
 #
@@ -573,6 +467,8 @@ def cast_to_qname(value, namespaces=None):
 
 
 def cast_to_datetime(value, tz=None):
+    if isinstance(value, DateTime10):
+        return value
     return DateTime10.fromstring(value, tzinfo=tz)
 
 
@@ -601,14 +497,16 @@ def nud(self):
 @method('QName')
 def evaluate(self, context=None):
     if self.label == 'constructor':
-        item = self.get_argument(context)
-        if item is None:
+        arg = self.data_value(self.get_argument(context))
+        if arg is None:
             return []
+
         try:
-            return self.cast(item, self.parser.namespaces)
+            if isinstance(arg, UntypedAtomic):
+                return self.cast(arg.value, self.parser.namespaces)
+            return self.cast(arg, self.parser.namespaces)
         except ElementPathError as err:
-            if err.token is None:
-                err.token = self
+            err.token = self
             raise
     else:
         uri = self.get_argument(context)
@@ -643,11 +541,14 @@ def evaluate(self, context=None):
 @method('dateTime')
 def evaluate(self, context=None):
     if self.label == 'constructor':
-        item = self.get_argument(context)
-        if item is None:
+        arg = self.data_value(self.get_argument(context))
+        if arg is None:
             return []
+
         try:
-            return self.cast(item, tz=None if context is None else context.timezone)
+            if isinstance(arg, UntypedAtomic):
+                return self.cast(arg.value, tz=None if context is None else context.timezone)
+            return self.cast(arg, tz=None if context is None else context.timezone)
         except ValueError as err:
             raise self.error('FOCA0002', str(err))
         except TypeError as err:
@@ -672,4 +573,15 @@ def cast(value):
     return UntypedAtomic(value)
 
 
+@method('untypedAtomic')
+def evaluate(self, context=None):
+    arg = self.data_value(self.get_argument(context))
+    if arg is None:
+        return []
+    elif isinstance(arg, UntypedAtomic):
+        return arg
+    else:
+        return self.cast(arg)
+
+
 XPath2Parser.build()  # XPath 2.0 definition complete, can build the parser class.
